from .account.role import RoleAsyncService, RoleSyncService
from .account.user import UserAsyncService, UserSyncService
from .organization.fixture import (
    FixtureAsyncService,
    FixtureSyncService,
    FixtureTeamAsyncService,
    FixtureTeamSyncService,
)
from .organization.league import LeagueAsyncService, LeagueSyncService
from .organization.manager import ManagerAsyncService, ManagerSyncService
from .organization.player import PlayerAsyncService, PlayerSyncService
from .organization.referee import RefereeAsyncService, RefereeSyncService
from .organization.schedule import ScheduleAsyncService, ScheduleSyncService
from .organization.season import SeasonAsyncService, SeasonSyncService
from .organization.standings import StandingsAsyncService, StandingsSyncService
from .organization.team import TeamAsyncService, TeamSyncService

__all__ = [
    "FixtureAsyncService",
    "FixtureSyncService",
    "FixtureTeamAsyncService",
    "FixtureTeamSyncService",
    "LeagueAsyncService",
    "LeagueSyncService",
    "ManagerAsyncService",
    "ManagerSyncService",
    "PlayerAsyncService",
    "PlayerSyncService",
    "RefereeAsyncService",
    "RefereeSyncService",
    "RoleAsyncService",
    "RoleSyncService",
    "ScheduleAsyncService",
    "ScheduleSyncService",
    "SeasonAsyncService",
    "SeasonSyncService",
    "StandingsAsyncService",
    "StandingsSyncService",
    "TeamAsyncService",
    "TeamSyncService",
    "UserAsyncService",
    "UserSyncService",
]



from datetime import UTC, datetime

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Boolean, Column, DateTime, String, Table
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import UUIDAuditBase, mapper, metadata

if TYPE_CHECKING:
    from .league import League


@define(slots=False)
class Season(UUIDAuditBase):
    """A season defines an overall period of time in which a league or leagues are active."""

    name: str | None = field(default=None, validator=validators.max_len(80))
    description: str | None = field(default=None, validator=validators.optional(validators.max_len(120)))
    active: bool = field(default=True)
    start_date: str | None = field(default=None)
    end_date: str | None = field(default=None)

    # leagues: list["League"] = field(default=list)


# SQLAlchemy Imperative Mappings

season = Table(
    "season",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column(
        "name",
        String(80),
        nullable=False,
        unique=True,
    ),
    Column("description", String(120), nullable=True),
    Column("active", Boolean, default=True),
    Column("start_date", String(10), nullable=True, default=None),
    Column("end_date", DateTime(), nullable=True, default=None),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)

# ORM Relationships

mapper.map_imperatively(
    Season,
    season,
    properties={
        "leagues": relationship("League", back_populates="season", lazy="selectin"),
        "schedules": relationship("Schedule", back_populates="season"),
    },
)



from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from attrs import define, field
from sqlalchemy import select

from leaguemanager.models import Referee, User
from leaguemanager.repository import RefereeAsyncRepository, RefereeSyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["RefereeSyncService", "RefereeAsyncService"]


class RefereeSyncService(SQLAlchemySyncRepositoryService):
    """Handles sync database operations for referees."""

    repository_type = RefereeSyncRepository


class RefereeAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles async database operations for referees."""

    repository_type = RefereeAsyncRepository



from advanced_alchemy.filters import CollectionFilter, OrderBy
from attrs import define, field
from sqlalchemy import select
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import Session

from leaguemanager import models
from leaguemanager.services import (
    FixtureSyncService,
    FixtureTeamSyncService,
    StandingsSyncService,
    TeamSyncService,
)


@define
class StandingsTable:
    """Responsible for creating and updating the standings table for a schedule."""

    session: Session
    schedule: models.Schedule

    standings_service: StandingsSyncService = field(init=False)
    fixture_service: FixtureSyncService = field(init=False)
    fixture_team_service: FixtureTeamSyncService = field(init=False)
    team_service: TeamSyncService = field(init=False)

    def __attrs_post_init__(self):
        # if not self.schedule.fixtures:
        #     raise ValueError("No fixtures found for schedule")
        self.standings_service = StandingsSyncService(session=self.session)
        self.fixture_service = FixtureSyncService(session=self.session)
        self.fixture_team_service = FixtureTeamSyncService(session=self.session)
        self.team_service = TeamSyncService(session=self.session)

    def get_or_create_current_standings(self):
        """Get or create the current standings for the schedule.

        If the schedule does not have any team standings, we create them.

        Returns:
            Sequence[models.Standings]: All standings (per team) for the schedule
        """
        if not self.schedule.team_standings:
            standings = []
            for team in self.schedule.league.teams:
                standings.append(
                    models.Standings(
                        schedule_id=self.schedule.id,
                        team_id=team.id,
                    )
                )
            standings = self.standings_service.create_many(standings)
            return standings
        return self.schedule.team_standings

    def link_teams_to_fixtures(self):
        """Link all fixtures for the schedule to their teams.

        This is done through an association table (fixture_team). For each fixture,
        we create two fixture_team rows, one for the home team and one for the away team.

        Raises:
            NoResultFound: In the event that no fixtures are found, we raise an exception

        Returns:
            Sequence[models.FixtureTeam]: List of association tables (fixture_team) with fixture data.
        """
        all_fixtures = self.fixture_service.all_fixtures_in_schedule(self.schedule.id)

        if not all_fixtures:
            raise NoResultFound(f"No fixtures found for {self.schedule}")

        for fixture in all_fixtures:
            home_fixture_team = models.FixtureTeam(
                fixture_id=fixture.id,
                team_id=fixture.home_team_id,
            )
            self.fixture_team_service.create(home_fixture_team)

            away_fixture_team = models.FixtureTeam(
                fixture_id=fixture.id,
                team_id=fixture.away_team_id,
            )
            self.fixture_team_service.create(away_fixture_team)

        return self.fixture_team_service.list()

    @property
    def teams(self):
        return self.schedule.league.teams

    @property
    def all_fixtures(self):
        """List of all fixtures in the schedule."""
        return self.fixture_service.all_fixtures_in_schedule(self.schedule.id)

    @property
    def past_fixtures(self):
        """List of fixtures that have already been played."""
        return self.fixture_service.past_fixtures(self.schedule.id)

    @property
    def remaining_fixtures(self):
        """List of fixtures that have not yet been played."""
        return self.fixture_service.remaining_fixtures(self.schedule.id)

    def next_scheduled_fixtures(self, days: int = 6, **kwargs):
        return self.fixture_service.next_scheduled_fixtures(self.schedule.id, days=days, **kwargs)

    def generate_standings_table(self, order_by_field: str = "points", sort_order: str = "desc"):
        """Retrieve standings for all teams.

        The order defaults to descending by points (first in order has most points), although
        any field can be used. In case of a tie, the order is by goals difference. (The highest
        goals difference is ordered first.)

        Args:
            order_by_field (str, optional): Field to order by. Defaults to "points".
            sort_order (str, optional): Either "asc" or "desc". Defaults to "desc".

        Returns:
            Sequence[models.Standings]: Standings table for all teams
        """

        standings = self.standings_service.list(
            CollectionFilter(field_name="team_id", values=[team.id for team in self.teams]),
            OrderBy(field_name=order_by_field, sort_order=sort_order),
            OrderBy(field_name="goals_diff", sort_order="desc"),
            statement=select(models.Standings).where(models.Standings.schedule_id == self.schedule.id),
        )
        return standings



from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as _UUID
from sqlalchemy import Boolean, Column, DateTime, Float, ForeignKey, String, Table
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import UUIDAuditBase, mapper, metadata

from ._association import fixture_referee

if TYPE_CHECKING:
    from .user import User


@define(slots=False)
class Referee(UUIDAuditBase):
    """A referee for a fixture."""

    user_id: UUID | None = field(default=None)
    fixture_id: UUID | None = field(default=None)
    verified: bool = field(default=False)
    first_name: str | None = field(default=None, validator=validators.max_len(40))
    last_name: str | None = field(default=None, validator=validators.max_len(40))
    note: str | None = field(default=None, validator=validators.optional(validators.max_len(255)))
    license: str | None = field(default=None, validator=validators.max_len(20))
    price_per_game: float | None = field(default=None)

    user_email: AssociationProxy[str] = association_proxy("user", "email")


# SQLalchemy Imperative Mapping

referee = Table(
    "referee",
    metadata,
    Column("id", _UUID, primary_key=True),
    Column("user_id", _UUID, ForeignKey("user.id")),
    Column("fixture_id", _UUID, ForeignKey("fixture.id")),
    Column("first_name", String(40)),
    Column("last_name", String(40)),
    Column("note", String(255)),
    Column("license", String(20)),
    Column("verified", Boolean),
    Column("price_per_game", Float),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)

# ORM Relationships

mapper.map_imperatively(
    Referee,
    referee,
    properties={
        "user": relationship("User", back_populates="referee", uselist=False),
        "fixtures": relationship("Fixture", secondary=fixture_referee, back_populates="referees"),
    },
)



import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from leaguemanager.db import register_sqlite
from leaguemanager.models.base import metadata
from leaguemanager.services import (
    FixtureSyncService,
    FixtureTeamSyncService,
    LeagueSyncService,
    ScheduleSyncService,
    SeasonSyncService,
    StandingsSyncService,
    TeamSyncService,
)

pytest_plugins = ["tests.data_fixtures"]

register_sqlite()


@pytest.fixture(scope="session")
def db_dir(tmp_path_factory):
    db_dir = tmp_path_factory.mktemp("db_test")
    return db_dir


# Sync DB fixtures


@pytest.fixture(scope="session")
def engine(db_dir):
    uri = f"sqlite:///{db_dir / 'test.db'}"
    engine = create_engine(uri, echo=False)
    metadata.drop_all(engine)
    metadata.create_all(engine)
    yield engine
    metadata.drop_all(engine)
    engine.dispose()


@pytest.fixture(scope="session")
def session_factory(engine):
    return sessionmaker(bind=engine, expire_on_commit=True, autocommit=False, autoflush=False)


@pytest.fixture(scope="session", autouse=True)
def session(session_factory, all_data):
    session_ = session_factory()
    yield session_
    session_.rollback()
    session_.close()


@pytest.fixture(scope="session")
def season_service(session):
    return SeasonSyncService(session=session)


@pytest.fixture(scope="session")
def team_service(session):
    return TeamSyncService(session=session)


@pytest.fixture(scope="session")
def league_service(session):
    return LeagueSyncService(session=session)


@pytest.fixture(scope="session")
def schedule_service(session):
    return ScheduleSyncService(session=session)


@pytest.fixture(scope="session")
def fixture_service(session):
    return FixtureSyncService(session=session)


@pytest.fixture(scope="session")
def fixture_team_service(session):
    return FixtureTeamSyncService(session=session)


@pytest.fixture(scope="session")
def standings_service(session):
    return StandingsSyncService(session=session)



from datetime import timedelta

from attrs import define, field
from sqlalchemy.orm import Session

from leaguemanager.models import Fixture, Schedule, Team
from leaguemanager.services import (
    FixtureSyncService,
    FixtureTeamSyncService,
    StandingsSyncService,
    TeamSyncService,
)

from ._generator import ScheduleGenerator

__all__ = ["Scheduler"]


@define
class Scheduler:
    session: Session
    schedule: Schedule
    generator: ScheduleGenerator = field(init=False)

    standings_service: StandingsSyncService = field(init=False)
    fixture_service: FixtureSyncService = field(init=False)
    fixture_team_service: FixtureTeamSyncService = field(init=False)
    team_service: TeamSyncService = field(init=False)

    def __attrs_post_init__(self):
        # if not self.schedule.fixtures:
        #     raise ValueError("No fixtures found for schedule")
        self.standings_service = StandingsSyncService(session=self.session)
        self.fixture_service = FixtureSyncService(session=self.session)
        self.fixture_team_service = FixtureTeamSyncService(session=self.session)
        self.team_service = TeamSyncService(session=self.session)
        self.generator = ScheduleGenerator(schedule=self.schedule)

    def sort_fixtures(self, order_by_field: str = "date", sort_order: str = "asc"):
        sorted_fixtures = sorted(
            self.schedule.fixtures,
            key=lambda fixture: getattr(fixture, order_by_field),
            reverse=(sort_order == "desc"),
        )
        return sorted_fixtures

    @property
    def last_fixture(self):
        return self.sort_fixtures(order_by_field="date", sort_order="desc")[0]

    def push_fixture_to_end_of_schedule(self, fixture: Fixture):
        matchday_diff = (self.last_fixture.matchday + 1) - fixture.matchday
        fixture.date = fixture.date + timedelta(days=matchday_diff * 7)
        return fixture

    def push_matchday_to_end_of_schedule(self, matchday: int):
        last_fixture_matchday = self.last_fixture.matchday
        for fixture in self.schedule.fixtures:
            if fixture.matchday == matchday:
                matchday_diff = last_fixture_matchday + 1 - fixture.matchday
                fixture.date = fixture.date + timedelta(days=matchday_diff * 7)

    def push_all_fixtures_by_one_week(self, fixture: Fixture):
        # Push all fixtures by one week with date after the fixture
        for f in self.schedule.fixtures:
            if f.matchday >= fixture.matchday:
                f.date = f.date + timedelta(days=7)
        return fixture

    def generate_fixtures(self, shuffle_order: bool = True, teams: list[Team] | None = None, **kwargs) -> list[Fixture]:
        """Generates fixtures for the entire season.

        Included here for convenience, but can be accessed through the `generator` attribute
        as well.

        Args:
            shuffle_order (bool, optional): Determines whether fixtures will be randomized. Defaults to True.
            teams (list[Team], optional): Teams specific to the schedule/league. Defaults to None.
        """
        generator = ScheduleGenerator(schedule=self.schedule)
        return generator.generate_fixtures(shuffle_order=shuffle_order, teams=teams, **kwargs)



from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field
from sqlalchemy import UUID as _UUID
from sqlalchemy import Column, DateTime, ForeignKey, Integer, Table, UniqueConstraint
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import UUIDAuditBase, mapper, metadata

if TYPE_CHECKING:
    from .league import League
    from .season import Season
    from .team import Team


@define(slots=False)
class Standings(UUIDAuditBase):
    """Standings table for a given team. The stats are geared toward soccer leagues."""

    schedule_id: UUID | None = field(default=None)
    team_id: UUID | None = field(default=None)

    games_played: int = field(default=0)
    games_won: int = field(default=0)
    games_drawn: int = field(default=0)
    games_lost: int = field(default=0)
    goals_for: int = field(default=0)
    goals_against: int = field(default=0)
    goals_diff: int = field(default=0)
    points: int = field(default=0)

    # team_name: AssociationProxy[str] | None = association_proxy("team", "name")


# SQLAlchemy Imperative Mapping

standings = Table(
    "standings",
    metadata,
    Column("id", _UUID, primary_key=True),
    Column("schedule_id", _UUID, ForeignKey("schedule.id")),
    Column("team_id", _UUID, ForeignKey("team.id")),
    # UniqueConstraint("id", "schedule_id", "team_id", name="uxstandings"),
    Column("games_played", Integer, nullable=False),
    Column("games_won", Integer, nullable=False),
    Column("games_drawn", Integer, nullable=False),
    Column("games_lost", Integer, nullable=False),
    Column("goals_for", Integer, nullable=False),
    Column("goals_against", Integer, nullable=False),
    Column("points", Integer, nullable=False),
    Column("goals_diff", Integer, nullable=False),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)

# ORM Relationships

mapper.map_imperatively(
    Standings,
    standings,
    properties={
        "schedule": relationship("Schedule", back_populates="team_standings", lazy="selectin"),
        "team": relationship("Team", back_populates="standings"),
    },
)



from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import League, Season
from leaguemanager.repository import LeagueAsyncRepository, LeagueSyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["LeagueSyncService", "LeagueAsyncService"]


class LeagueSyncService(SQLAlchemySyncRepositoryService):
    """Handles sync database operations for a league."""

    repository_type = LeagueSyncRepository

    def get_active_leagues(self, *filters: FilterTypes, **kwargs) -> League:
        """Get active leagues."""

        statement = select(League).where(League.active == True)  # noqa

        return self.list(*filters, statement=statement, **kwargs)

    def get_leagues_in_season(self, *filters: FilterTypes, season_id: UUID, **kwargs) -> Season:
        """Get the current season."""

        statement = select(League).join_from(Season, League).where(Season.id == season_id)

        return self.list(*filters, statement=statement, **kwargs)


class LeagueAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles async database operations for a league."""

    repository_type = LeagueAsyncRepository

    async def get_active_leagues(self, *filters: FilterTypes, **kwargs) -> League:
        """Get active leagues."""

        statement = select(League).where(League.active == True)  # noqa

        return await self.list(*filters, statement=statement, **kwargs)



from datetime import datetime, timedelta
from uuid import uuid4

import pytest
from advanced_alchemy.filters import CollectionFilter, OnBeforeAfter
from sqlalchemy.exc import NoResultFound

from leaguemanager.core.toolbox import clear_table
from leaguemanager.domain import Scheduler, StandingsTable, TabulateFixture
from leaguemanager.models import Fixture, FixtureTeam, Schedule, Standings, Team
from leaguemanager.services import (
    FixtureSyncService,
    LeagueSyncService,
    ScheduleSyncService,
    SeasonSyncService,
    TeamSyncService,
)


@pytest.fixture(scope="module")
def season(season_service) -> SeasonSyncService:
    season = season_service.create(
        data={"name": "SZNZ", "description": "Winter", "start_date": "2022-01-01"}, auto_commit=True
    )
    return season


@pytest.fixture(scope="module")
def league(league_service, season) -> LeagueSyncService:
    league = league_service.create(data={"name": "Men's 7v7", "category": "MEN", "division": "A", "match_day": "SUN"})
    league.season = season
    league = league_service.update(league, auto_commit=True)
    return league


@pytest.fixture(scope="module")
def teams(team_service, league_service, league) -> TeamSyncService:
    teams = team_service.create_many(
        [
            {"name": "Team 1", "league_id": league.id},
            {"name": "Team 2", "league_id": league.id},
            {"name": "Team 3", "league_id": league.id},
            {"name": "Team 4", "league_id": league.id},
        ]
    )
    league.teams = teams
    league_service.update(league, auto_commit=True)
    return teams


@pytest.fixture(scope="module")
def schedule(schedule_service, league, season, teams) -> ScheduleSyncService:
    _schedule = {
        "name": "I Want A Dog",
        "total_games": 10,
        "game_length": 45,
        "time_between_games": 15,
        "start_date": "2025-1-05 16:00:00",
        "season_id": season.id,
        "league_id": league.id,
        "teams": teams,
    }

    schedule = schedule_service.create(_schedule)
    return schedule


@pytest.fixture(scope="module")
def scheduler(session, schedule) -> Scheduler:
    scheduler = Scheduler(session=session, schedule=schedule)
    return scheduler


@pytest.fixture(scope="module")
def league_fixtures(fixture_service, scheduler) -> FixtureSyncService:
    fixtures = scheduler.generate_fixtures(shuffle_teams=False)
    fixture_service.create_many(fixtures, auto_commit=True)
    return fixtures


@pytest.fixture(scope="module")
def standings_table(session, schedule):
    standings_table = StandingsTable(
        session=session,
        schedule=schedule,
    )

    return standings_table


def test_standings_table_properties(schedule, standings_table):
    assert standings_table.schedule == schedule
    assert standings_table.schedule.season == schedule.season
    assert standings_table.schedule.season.name == "SZNZ"


def test_get_or_create_standings(standings_table, standings_service):
    standings = standings_table.get_or_create_current_standings()
    assert len(standings) == 4

    for standing in standings:
        standing.games_played = 1
        standings_service.update(standing, auto_commit=True)

    new_standings = standings_table.get_or_create_current_standings()
    assert len(new_standings) == 4
    assert new_standings[0].games_played == 1
    assert new_standings[1].games_played == 1
    assert new_standings[2].games_played == 1
    assert new_standings[3].games_played == 1


def test_past_fixtures_all_unplayed(standings_table, league_fixtures, fixture_service):
    fixtures = standings_table.past_fixtures
    assert len(fixtures) == 0

    _fixture = league_fixtures[0]
    _fixture.status = "P"
    fixture_service.update(_fixture, auto_commit=True)

    fixtures = standings_table.past_fixtures
    assert len(fixtures) == 1
    _fixture = league_fixtures[0]
    _fixture.status = "U"
    fixture_service.update(_fixture, auto_commit=True)


def test_remaining_fixtures_all_unplayed(standings_table, league_fixtures, fixture_service):
    fixtures = standings_table.remaining_fixtures
    assert len(fixtures) == 16

    _fixture = league_fixtures[0]
    _fixture.status = "P"
    fixture_service.update(_fixture, auto_commit=True)

    fixtures = standings_table.remaining_fixtures
    assert len(fixtures) == 15
    _fixture = league_fixtures[0]
    _fixture.status = "U"
    fixture_service.update(_fixture, auto_commit=True)


@pytest.mark.parametrize("time_delta_days, expected", [(6, 2), (12, 4), (24, 8)])
def test_next_scheduled_fixtures(standings_table, league_fixtures, time_delta_days, expected):
    # Note: league_fixtures needs to be here to generate fixtures for this test
    # if run individually.
    all_fixtures = standings_table.all_fixtures
    assert len(all_fixtures) == 16
    assert all_fixtures[0].date == datetime(2025, 1, 5, 16, 0)
    assert all_fixtures[1].date == datetime(2025, 1, 5, 16, 0)

    filter = OnBeforeAfter(
        field_name="date",
        on_or_after=datetime(2025, 1, 4, 16, 0),
        on_or_before=datetime(2025, 1, 4, 16, 0) + timedelta(days=time_delta_days),
    )

    fixtures = standings_table.next_scheduled_fixtures(filter=filter)
    assert len(fixtures) == expected


def test_link_teams_to_fixtures(standings_table, league_fixtures):
    # Note: league_fixtures needs to be here to generate fixtures for this test
    # if run individually.
    fixture_teams = standings_table.link_teams_to_fixtures()
    assert len(fixture_teams) == 32


def test_link_teams_to_fixtures_when_no_fixtures(session, schedule_service):
    temp_schedule = Schedule(
        season_id=uuid4(),
        league_id=uuid4(),
        start_date="2021-1-08 18:00:00",
        end_date="2022-02-01 18:00:00",
        total_games=0,
    )
    temp_standings = StandingsTable(session=session, schedule=temp_schedule)

    with pytest.raises(NoResultFound):
        temp_standings.link_teams_to_fixtures()


def test_team_update_standings(
    league, session, schedule, standings_service, fixture_team_service, mocker, team_service
):
    mock_home_team = Team(league_id=league.id, name="Mock Home")
    mock_away_team = Team(league_id=league.id, name="Mock Away")

    team_service.create(mock_home_team)
    team_service.create(mock_away_team)

    fake_fixture = Fixture(
        home_team_id=mock_home_team.id,
        away_team_id=mock_away_team.id,
        schedule_id=uuid4(),
        status="P",
        home_goals=3,
        away_goals=1,
    )

    mock_home_stat = FixtureTeam(
        team_id=mock_home_team.id,
        fixture_id=fake_fixture.id,
        is_played=True,
        points=3,
        goals_for=3,
        goals_against=1,
        fixture_result="W",
    )
    mock_away_stat = FixtureTeam(
        team_id=mock_away_team.id,
        fixture_id=fake_fixture.id,
        is_played=True,
        points=0,
        goals_for=1,
        goals_against=3,
        fixture_result="L",
    )

    fixture_team_service.create(mock_home_stat)
    fixture_team_service.create(mock_away_stat)

    mock_standings = Standings(
        schedule_id=schedule.id,
        team_id=mock_home_team.id,
    )

    standings_service.create(mock_standings)

    tabulate_fixture = TabulateFixture(
        session=session,
        fixture=fake_fixture,
    )

    expected_standings_after_update = Standings(
        schedule_id=schedule.id,
        team_id=mock_home_team.id,
        points=3,
        games_played=1,
        games_won=1,
        games_drawn=0,
        games_lost=0,
        goals_for=3,
        goals_against=1,
        goals_diff=2,
    )

    mocker.patch.object(standings_service, "get", return_value=mock_standings)

    updated_standings = tabulate_fixture.update_standings(mock_home_team, mock_home_stat)
    assert updated_standings.points == expected_standings_after_update.points
    assert updated_standings.goals_for == expected_standings_after_update.goals_for
    assert updated_standings.games_played == expected_standings_after_update.games_played

    team_service.delete(mock_home_team.id)
    team_service.delete(mock_away_team.id)
    fixture_team_service.delete(mock_home_stat.id)
    fixture_team_service.delete(mock_away_stat.id)
    standings_service.delete(mock_standings.id)


@pytest.mark.parametrize(["prior_home_points", "expected"], [(0, 3), (1, 4), (6, 9)])
def test_process_fixture_results(
    league, schedule, session, standings_service, fixture_team_service, team_service, prior_home_points, expected
):
    mock_home_team = Team(league_id=league.id, name="Mock Home")
    mock_away_team = Team(league_id=league.id, name="Mock Away")

    team_service.create(mock_home_team)
    team_service.create(mock_away_team)

    fake_fixture = Fixture(
        home_team_id=mock_home_team.id,
        away_team_id=mock_away_team.id,
        schedule_id=uuid4(),
        status="P",
        home_goals=3,
        away_goals=1,
    )

    mock_home_stat = FixtureTeam(
        team_id=mock_home_team.id,
        fixture_id=fake_fixture.id,
        is_played=True,
        points=3,
        goals_for=3,
        goals_against=1,
        fixture_result="W",
    )
    mock_away_stat = FixtureTeam(
        team_id=mock_away_team.id,
        fixture_id=fake_fixture.id,
        is_played=True,
        points=0,
        goals_for=1,
        goals_against=3,
        fixture_result="L",
    )

    fixture_team_service.create(mock_home_stat)
    fixture_team_service.create(mock_away_stat)

    home_standings = Standings(
        schedule_id=schedule.id,
        team_id=mock_home_team.id,
        points=prior_home_points,
    )

    away_standings = Standings(
        schedule_id=schedule.id,
        team_id=mock_away_team.id,
    )

    standings_service.create(home_standings)
    standings_service.create(away_standings)

    tabulate_fixture = TabulateFixture(session=session, fixture=fake_fixture)

    home_standings, away_standings = tabulate_fixture.process_fixture_results()

    assert home_standings.games_played == 1
    assert home_standings.games_won == 1
    assert home_standings.games_drawn == 0
    assert home_standings.games_lost == 0
    assert home_standings.goals_for == 3
    assert home_standings.goals_against == 1
    assert home_standings.goals_diff == 2
    assert home_standings.points == expected

    assert away_standings.games_played == 1
    assert away_standings.games_won == 0
    assert away_standings.games_drawn == 0
    assert away_standings.games_lost == 1
    assert away_standings.goals_for == 1
    assert away_standings.goals_against == 3
    assert away_standings.goals_diff == -2
    assert away_standings.points == 0

    team_service.delete(mock_home_team.id)
    team_service.delete(mock_away_team.id)
    fixture_team_service.delete(mock_home_stat.id)
    fixture_team_service.delete(mock_away_stat.id)
    standings_service.delete(home_standings.id)
    standings_service.delete(away_standings.id)


def test_generate_standings_table(schedule, session, standings_service, standings_table):
    standings_1 = Standings(
        schedule_id=schedule.id,
        team_id=standings_table.teams[0].id,
        points=3,
        games_played=1,
        games_won=1,
        games_drawn=0,
        games_lost=0,
        goals_for=3,
        goals_against=1,
        goals_diff=-5,
    )

    standings_2 = Standings(
        schedule_id=schedule.id,
        team_id=standings_table.teams[1].id,
        points=3,
        games_played=1,
        games_won=1,
        games_drawn=0,
        games_lost=0,
        goals_for=3,
        goals_against=1,
        goals_diff=2,
    )

    standings_3 = Standings(
        schedule_id=schedule.id,
        team_id=standings_table.teams[2].id,
        points=6,
        games_played=1,
        games_won=1,
        games_drawn=0,
        games_lost=0,
        goals_for=3,
        goals_against=1,
        goals_diff=2,
    )

    standings_4 = Standings(
        schedule_id=schedule.id,
        team_id=standings_table.teams[3].id,
        points=12,
        games_played=1,
        games_won=1,
        games_drawn=0,
        games_lost=0,
        goals_for=3,
        goals_against=1,
        goals_diff=2,
    )

    standings_other_schedule = Standings(
        schedule_id=uuid4(),
        team_id=standings_table.teams[3].id,
        points=200,
        games_played=1,
        games_won=1,
        games_drawn=0,
        games_lost=0,
        goals_for=3,
        goals_against=1,
        goals_diff=2,
    )

    standings_other_team = Standings(
        schedule_id=schedule.id,
        team_id=uuid4(),
        points=100,
        games_played=1,
        games_won=1,
        games_drawn=0,
        games_lost=0,
        goals_for=3,
        goals_against=1,
        goals_diff=2,
    )

    clear_table(session, Standings)
    standings_service.create_many(
        [standings_1, standings_2, standings_3, standings_4, standings_other_schedule, standings_other_team]
    )

    standings = standings_table.generate_standings_table()

    assert len(standings) == 4
    assert standings[0].team_id == standings_table.teams[3].id
    assert standings[0].points == 12
    assert standings[1].team_id == standings_table.teams[2].id
    assert standings[2].team_id == standings_table.teams[1].id
    assert standings[2].points == 3
    assert standings[2].goals_diff == 2
    assert standings[3].team_id == standings_table.teams[0].id
    assert standings[3].points == 3
    assert standings[3].goals_diff == -5

    clear_table(session, Standings)



from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import Manager
from leaguemanager.repository import ManagerAsyncRepository, ManagerSyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["ManagerSyncService", "ManagerAsyncService"]


class ManagerSyncService(SQLAlchemySyncRepositoryService):
    """Handles sync database operations for managers."""

    repository_type = ManagerSyncRepository


class ManagerAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for managers."""

    repository_type = ManagerAsyncRepository



from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import League, Team
from leaguemanager.repository import TeamAsyncRepository, TeamSyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["TeamSyncService", "TeamAsyncService"]


class TeamSyncService(SQLAlchemySyncRepositoryService):
    """Handles sync database operations for team data."""

    repository_type = TeamSyncRepository


class TeamAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles async database operations for team data."""

    repository_type = TeamAsyncRepository



import datetime
from collections import deque
from itertools import chain
from random import shuffle
from typing import Any, Iterable

from attrs import define

from leaguemanager.models import Fixture, Schedule, Team
from leaguemanager.models.enums import Field

__all__ = ["ScheduleGenerator"]


@define
class ScheduleGenerator:
    """Schedules fixtures(matches) based on the league configuration."""

    schedule: Schedule
    blackout_dates: list[datetime.datetime] | None = None  # TODO: Implement with days or ranges?

    @property
    def total_games(self) -> int:
        return self.schedule.total_games

    @property
    def teams(self) -> list[Team]:
        try:
            return self.schedule.league.teams
        except AttributeError:
            raise ValueError("The league does not have any teams associated.") from AttributeError

    def split_teams(self, teams: Iterable[Team]) -> tuple[list[Team], list[Team]]:
        """Split a list of teams in half, raising an error if the list is odd."""

        if len(teams) % 2 != 0:
            raise ValueError("List length must be even.")

        return (
            teams[: len(teams) // 2],
            teams[len(teams) // 2 :],
        )

    def determine_field(self, match_count: int, concurrent_games: int | None = None) -> str:
        fields = [f.name for f in Field]
        if not concurrent_games:
            concurrent_games = self.schedule.concurrent_games
        if match_count == 1 or match_count > concurrent_games:
            return fields[0]
        return fields[match_count - 1]

    def home_or_away(self, round_number: int, match: tuple[Team, Team]) -> tuple[Team, Team]:
        if round_number % 2 == 0:
            home_team, away_team = match
        else:
            home_team, away_team = match[::-1]

        return home_team, away_team

    def check_for_bye(self, match: tuple[Team, Team]) -> bool:
        if match[0].name == "Bye Week" or match[1].name == "Bye Week":
            return True
        return False

    def increment_matchday(self, matchday: int, *, increment_by_days: int = 7) -> None:
        matchday_start_time = self.schedule.start_date
        if matchday > 1:
            matchday_start_time += datetime.timedelta(days=increment_by_days * (matchday - 1))
        return matchday_start_time

    def determine_start_time(self, matchday: int, match_count: int) -> datetime.datetime:
        game_date = self.increment_matchday(matchday)

        if match_count <= self.schedule.concurrent_games:
            return game_date
        increment_by = self.schedule.game_length + self.schedule.half_time + self.schedule.time_between_games
        multiplier = (match_count - 1) // self.schedule.concurrent_games
        minute_increment = increment_by * (multiplier)
        return game_date + datetime.timedelta(minutes=minute_increment)

    def create_matchups(self, matchday: int, teams: list[Team] | None = None) -> list[tuple[Team]]:
        """Creates matchups by rotating teams based on the matchday.

        See: https://en.wikipedia.org/wiki/Round-robin_tournament#Circle_method for
        details on round-robin scheduling. In order to account for "odd" number of teams,
        we add a temporary "bye week" team to the list of teams, anchoring the first team
        for rotation.

        Args:
            matchday (int): Current matchday
            teams (list[Team]) | None: List of teams. Defaults to self.teams

        Returns:
            matchups (list[tuple[Team]]): List of tuples corresponding to matchups
        """
        if not teams:
            teams = self.teams
        if len(teams) % 2 != 0:
            teams.append(Team(name="Bye Week"))
        home, away = self.split_teams(teams)
        matchups = list(zip(home, away, strict=True))
        if not matchday == 1:
            _teams = deque(chain(*matchups))
            anchor = _teams.popleft()
            _teams.rotate(matchday)
            _teams.appendleft(anchor)
            home, away = self.split_teams([*_teams])
            matchups = list(zip(home, away, strict=True))
        return matchups

    def create_matchday_fixtures(
        self, matchday: int, round_number: int, teams: list[Team] | None = None
    ) -> list[Fixture]:
        """Creates fixtures for a single matchday.

        Args:
            matchday (int): Current matchday
            round_number (int): Current round number
            teams (list[Team]) | None: List of teams. Defaults to self.teams

        Returns:
            fixtures (list[Fixture]): List of fixtures

        """
        fixtures = []
        matchups = self.create_matchups(matchday, teams)

        for match_count, match in enumerate(matchups, start=1):
            home, away = self.home_or_away(round_number=round_number, match=match)
            start_time = self.determine_start_time(matchday=matchday, match_count=match_count)
            field = self.determine_field(match_count=match_count)

            fixture = Fixture(
                schedule_id=self.schedule.id,
                season_id=self.schedule.season_id,
                league_id=self.schedule.league_id,
                home_team_id=home.id,
                away_team_id=away.id,
                bye_week=self.check_for_bye(match),
                matchday=matchday,
                date=start_time,
                field=field,
            )

            fixtures.append(fixture)

        return fixtures

    def generate_fixtures(self, shuffle_order: bool = True, teams: list[Team] | None = None, **kwargs) -> list[Any]:
        """Generates fixtures for the entire season."""
        if not teams:
            teams = self.teams
        if self.schedule.fixtures:
            raise ValueError("Schedule already has fixtures...")
        all_fixtures = []
        if shuffle_order:
            shuffle(teams)

        # Calculates number of full rounds, and remaining games (if any) left in the schedule
        rounds, remaining = divmod(self.total_games, len(teams))

        if rounds == 0:
            raise ValueError("Not enough matchdays for all teams to play each other.")

        for matchday in range(1, (self.total_games - remaining) + 1):
            round_number = matchday // (len(teams))
            round_number += 1
            matchday_fixtures = self.create_matchday_fixtures(matchday, round_number, teams)
            all_fixtures.extend(matchday_fixtures)

        return all_fixtures

    def generate_tournament_fixtures(self) -> None:
        raise NotImplementedError("Not implemented yet")



from .account.role import Role
from .account.user import User
from .account.user_role import UserRole
from .organization.fixture import Fixture
from .organization.fixture_team import FixtureTeam
from .organization.league import League
from .organization.manager import Manager
from .organization.player import Player
from .organization.referee import Referee
from .organization.schedule import Schedule
from .organization.season import Season
from .organization.standings import Standings
from .organization.team import Team

__all__ = [
    "Fixture",
    "FixtureTeam",
    "League",
    "Manager",
    "Player",
    "Referee",
    "Role",
    "Schedule",
    "Season",
    "Standings",
    "Team",
    "User",
    "UserRole",
]



from contextlib import asynccontextmanager, contextmanager
from typing import Any, AsyncGenerator, Generator

from attrs import define, field
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session
from svcs import Container, Registry

from leaguemanager.core import get_settings, toolbox
from leaguemanager.db import async_config, get_async_session, get_session, sync_config
from leaguemanager.dependency.managers import RepositoryManagement, ServiceManagement
from leaguemanager.services._typing import (
    AsyncRepositoryT,
    AsyncServiceT,
    SQLAlchemyAsyncConfigT,
    SQLAlchemySyncConfigT,
    SyncRepositoryT,
    SyncServiceT,
)

__all__ = ["LeagueManagerRegistry"]

settings = get_settings()


@define
class LeagueManagerRegistry:
    """Registry for managing services.

    TODO: Serve up async repos/services

    If no `Registry` is provided, this class creates one. Keep in mind that there should only
    be one registry per application. Database session is kept in a `Container` and is provided
    as needed. The same is true for all League Manager "services", which consist of database
    operations on those specific database tables (corresponding to its model).

    Attributes:
        service_registry (Registry | None): An `svcs` Registry for managing services.
        get_session (Generator[Session, Any, None]): A generator for a database session.
        get_async_session (AsyncGenerator[AsyncSession, Any]): A generator for an async database session.
        manager_sync_repos (list[type[SyncRepositoryT]]): List of repositories for sync database operations.
        manager_async_repos (list[type[AsyncRepositoryT]]): List of repositories for async database operations.
        manager_sync_services (list[type[SyncServiceT]]): List of services for sync database operations.
        manager_async_services (list[type[AsyncServiceT]]): List of services for async database operations.
        advanced_alchemy_sync_config (list[type[SQLAlchemySyncConfigT]]): List of Advanced Alchemy sync configurations.
        advanced_alchemy_async_config (list[type[SQLAlchemyAsyncConfigT]]): List of Advanced Alchemy async configurations.

    Example:
        >>> registry = LeagueManagerRegistry()
        >>> season_service = registry.provide_db_service(SeasonSyncService)
        >>> team_service = registry.provide_db_service(TeamSyncService)
        >>>
        >>> season_service.list()  #  List all seasons
        >>> team_service.count()  #  Count number of teams

    """

    service_registry: Registry | None = field(default=None)

    get_session: Generator[Session, Any, None] = field(default=get_session)
    get_async_session: AsyncGenerator[AsyncSession, Any] = field(default=get_async_session)

    manager_sync_repos: list[type[SyncRepositoryT]] = field()
    manager_async_repos: list[type[AsyncRepositoryT]] = field()
    manager_sync_services: list[type[SyncServiceT]] = field()
    manager_async_services: list[type[AsyncServiceT]] = field()

    aa_config_dir: str = field(default=settings.ADVANCED_ALCHEMY_CONFIG_DIRNAME)

    advanced_alchemy_sync_config: list[type[SQLAlchemySyncConfigT]] = field(init=False)
    advanced_alchemy_async_config: list[type[SQLAlchemyAsyncConfigT]] = field(init=False)

    # Useful for integrating with other apps/frameworks
    _SVCS_KEY_REGISTRY: str = field(default="league_manager_registry")
    _SVCS_KEY_CONTAINER: str = field(default="league_manager")

    @manager_sync_repos.default
    def _dynamic_sync_repos(self) -> list[type[SyncRepositoryT]]:
        return toolbox.get_repositories()

    @manager_async_repos.default
    def _dynamic_async_repos(self) -> list[type[AsyncRepositoryT]]:
        return toolbox.get_repositories(is_async=True)

    @manager_sync_services.default
    def _dynamic_sync_services(self) -> list[type[SyncServiceT]]:
        return toolbox.get_services()

    @manager_async_services.default
    def _dynamic_async_services(self) -> list[type[AsyncServiceT]]:
        return toolbox.get_services(is_async=True)

    def __attrs_post_init__(self):
        # check existing project has a python module not in venv or tests
        if not self.aa_config_dir:
            raise ValueError(
                "Cannot find a python module. Please create a .py file in the root or subdirectory of your project."
            )

        if not self.service_registry:
            self.service_registry = Registry()

        # Register Advanced Alchemy objects

        for repo_type in self.manager_sync_repos:
            self.register_db_repository(repository_type=repo_type)
        for service_type in self.manager_sync_services:
            self.register_db_service(service_type=service_type)

        if _dirname := self.aa_config_dir:
            _sync_config = toolbox.get_advanced_alchemy_config(_dirname)
            _async_config = toolbox.get_advanced_alchemy_config(_dirname, is_async=True)
        if not _sync_config:
            _sync_config = [sync_config]
        if not _async_config:
            _async_config = [async_config]
        for _config in _sync_config:
            self.registry.register_value(SQLAlchemySyncConfigT, _config)
        for _config in _async_config:
            self.registry.register_value(SQLAlchemyAsyncConfigT, _config)

    @property
    def registry(self) -> Registry:
        return self.service_registry

    # Register additional objects

    @contextmanager
    def sync_session_container(self) -> Generator[Container, None, None]:
        """Create a container for a sync database session."""
        self.registry.register_factory(Session, self.get_session)
        with Container(self.registry) as container:
            yield container

    @asynccontextmanager
    async def async_session_container(self) -> AsyncGenerator[Container, None]:
        """Create a container for an async database session."""
        self.registry.register_factory(AsyncSession, self.get_async_session)
        async with Container(self.registry) as container:
            yield container

    def register_db_repository(self, repository_type: type[SyncRepositoryT]) -> None:
        """Register a League Manager repository based on its type."""
        _repository = RepositoryManagement(repository_type=repository_type, db_session=self.provide_db_session)
        self.registry.register_value(repository_type, _repository.get_repository)

    def register_db_service(self, service_type: type[SyncServiceT]) -> None:
        """Register a League Manager service based on its type."""
        _service = ServiceManagement(service_type=service_type, db_session=self.provide_db_session)
        self.registry.register_value(service_type, next(_service.get_service))

    # # Retrieve objects

    @property
    def provide_db_session(self) -> Session:
        """Provide a sync database session."""
        with self.sync_session_container() as container:
            return container.get(Session)

    @property
    def provide_async_db_session(self) -> AsyncSession:
        """Provide an async database session."""
        with self.async_session_container() as container:
            return container.get(AsyncSession)

    @property
    def provide_sync_config(self) -> SQLAlchemySyncConfigT:
        return Container(self.registry).get(SQLAlchemySyncConfigT)

    @property
    def provide_async_config(self) -> SQLAlchemyAsyncConfigT:
        return Container(self.registry).get(SQLAlchemyAsyncConfigT)

    def provide_db_repository(self, repository_type: type[SyncRepositoryT]) -> type[SyncRepositoryT]:
        """Provide a League Manager repository based on its type."""
        return Container(self.registry).get(repository_type)

    def provide_db_service(self, service_type: type[SyncServiceT]) -> type[SyncServiceT]:
        """Provide a League Manager service based on its type."""
        return Container(self.registry).get(service_type)



from attrs import define, field
from sqlalchemy import select
from sqlalchemy.orm import Session

from leaguemanager import models
from leaguemanager.services import FixtureTeamSyncService, StandingsSyncService, TeamSyncService


@define
class TabulateFixture:
    """Updates the stats to the Fixture and FixtureTeam tables, as well as the standings."""

    session: Session
    fixture: models.Fixture

    standings_service: StandingsSyncService = field(init=False)
    fixture_team_service: FixtureTeamSyncService = field(init=False)
    team_service: TeamSyncService = field(init=False)

    def __attrs_post_init__(self):
        self.standings_service = StandingsSyncService(session=self.session)
        self.team_service = TeamSyncService(session=self.session)
        self.fixture_team_service = FixtureTeamSyncService(session=self.session)

    @property
    def home_team(self):
        """Get the current standings for the home team."""
        return self.team_service.get(self.fixture.home_team_id)

    @property
    def away_team(self):
        """Get the current standings for the away team."""
        return self.team_service.get(self.fixture.away_team_id)

    def process_fixture_results(self):
        """Updates the fixture_team tables with the results from the fixture.

        Note: These are currently tailored toward soccer leagues, obtaining 3 points per win and
        1 point per draw.
        """

        home_fixture_team = self.fixture_team_service.get(
            self.fixture.id,
            id_attribute="fixture_id",
            statement=select(models.FixtureTeam).where(models.FixtureTeam.team_id == self.fixture.home_team_id),
        )
        away_fixture_team = self.fixture_team_service.get(
            self.fixture.id,
            id_attribute="fixture_id",
            statement=select(models.FixtureTeam).where(models.FixtureTeam.team_id == self.fixture.away_team_id),
        )

        match self.fixture:
            case fixture if self.fixture.status in ["U", "D"]:
                raise ValueError(f"Invalid fixture status: {fixture.status}")
            case fixture if self.fixture.home_goals > fixture.away_goals:
                home_fixture_team.points = 3
                away_fixture_team.points = 0
                home_fixture_team.result = "W"
                away_fixture_team.result = "L"
            case fixture if self.fixture.home_goals < fixture.away_goals:
                home_fixture_team.points = 0
                away_fixture_team.points = 3
                home_fixture_team.result = "L"
                away_fixture_team.result = "W"
            case fixture if self.fixture.home_goals == fixture.away_goals:
                home_fixture_team.points = 1
                away_fixture_team.points = 1
                home_fixture_team.result = "D"
                away_fixture_team.result = "D"
            case fixture if self.fixture.status in ["P", "F", "A"]:
                home_fixture_team.is_played = True
                away_fixture_team.is_played = True

        home_fixture_team.goals_for = self.fixture.home_goals
        home_fixture_team.goals_against = self.fixture.away_goals
        away_fixture_team.goals_for = self.fixture.away_goals
        away_fixture_team.goals_against = self.fixture.home_goals

        home_fixture_team = self.fixture_team_service.update(home_fixture_team, auto_commit=True)
        away_fixture_team = self.fixture_team_service.update(away_fixture_team, auto_commit=True)

        home_standings = self.update_standings(self.home_team, home_fixture_team)
        away_standings = self.update_standings(self.away_team, away_fixture_team)

        return home_standings, away_standings

    def update_standings(self, team: models.Team, fixture_team: models.FixtureTeam):
        """Update the standings for the team.

        Args:
            team (models.Team): The team to update.
            fixture_team (models.FixtureTeam): The fixture_team to update.

        Returns:
            models.Standings: The updated standings.
        """
        _standings = self.standings_service.get(team.id, id_attribute="team_id")

        match fixture_team:
            case fixture_team if fixture_team.fixture_result == "W":
                _standings.games_won += 1
            case fixture_team if fixture_team.fixture_result == "L":
                _standings.games_lost += 1
            case fixture_team if fixture_team.fixture_result == "D":
                _standings.games_drawn += 1

        _standings.games_played += 1
        _standings.goals_for += fixture_team.goals_for
        _standings.goals_against += fixture_team.goals_against
        _standings.points += fixture_team.points
        _standings.goals_diff = _standings.goals_for - _standings.goals_against

        return self.standings_service.update(_standings, auto_commit=True)



from __future__ import annotations

from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as _UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, String, Table
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import UUIDAuditBase, mapper, metadata

from ._association import player_team

if TYPE_CHECKING:
    from ..account.user import User
    from .league import League
    from .player import Player
    from .standings import Standings


@define(slots=False)
class Team(UUIDAuditBase):
    """A team belonging to a specific league."""

    manager_id: UUID | None = field(default=None)
    league_id: UUID | None = field(default=None)
    active: bool = field(default=True)

    name: str | None = field(default=None, validator=validators.max_len(80))
    color_main: str | None = field(default=None, validator=validators.optional(validators.max_len(20)))
    color_secondary: str | None = field(default=None, validator=validators.optional(validators.max_len(20)))
    team_image: str | None = field(default=None)
    team_logo: str | None = field(default=None)

    # player_ids: list[AssociationProxy[str]] | None = association_proxy("players", "first_name")


# SQLAlchemy Imperative Mapping

team = Table(
    "team",
    metadata,
    Column("id", _UUID, primary_key=True),
    Column("manager_id", _UUID, ForeignKey("manager.id")),
    Column("league_id", _UUID, ForeignKey("league.id")),
    Column("name", String(80), nullable=False),
    Column("color_main", String(20)),
    Column("color_secondary", String(20)),
    Column("active", Boolean),
    Column("team_image", String(255)),
    Column("team_logo", String(255)),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)

# ORM Relationships

mapper.map_imperatively(
    Team,
    team,
    properties={
        "manager": relationship("Manager", back_populates="teams"),
        "league": relationship("League", back_populates="teams"),
        "players": relationship("Player", secondary=player_team, back_populates="teams"),
        "standings": relationship("Standings", back_populates="team"),
        "fixtures": relationship("FixtureTeam", back_populates="team"),
    },
)



from advanced_alchemy.repository import (
    SQLAlchemyAsyncRepository,
    SQLAlchemyAsyncSlugRepository,
)

from leaguemanager import models

__all__ = [
    "FixtureTeamAsyncRepository",
    "FixtureAsyncRepository",
    "LeagueAsyncRepository",
    "ManagerAsyncRepository",
    "PlayerAsyncRepository",
    "RefereeAsyncRepository",
    "RoleAsyncRepository",
    "ScheduleAsyncRepository",
    "SeasonAsyncRepository",
    "StandingsAsyncRepository",
    "TeamAsyncRepository",
    "UserAsyncRepository",
    "UserRoleAsyncRepository",
]


class FixtureTeamAsyncRepository(SQLAlchemyAsyncRepository[models.FixtureTeam]):
    """FixtureTeam repository."""

    model_type = models.FixtureTeam


class FixtureAsyncRepository(SQLAlchemyAsyncRepository[models.Fixture]):
    """Fixture repository."""

    model_type = models.Fixture


class LeagueAsyncRepository(SQLAlchemyAsyncSlugRepository[models.League]):
    """League repository."""

    model_type = models.League


class ManagerAsyncRepository(SQLAlchemyAsyncRepository[models.Manager]):
    """Manager repository."""

    model_type = models.Manager


class PlayerAsyncRepository(SQLAlchemyAsyncRepository[models.Player]):
    """Player repository."""

    model_type = models.Player


class RefereeAsyncRepository(SQLAlchemyAsyncRepository[models.Referee]):
    """Referee repository."""

    model_type = models.Referee


class RoleAsyncRepository(SQLAlchemyAsyncSlugRepository[models.Role]):
    """Role repository."""

    model_type = models.Role


class ScheduleAsyncRepository(SQLAlchemyAsyncRepository[models.Schedule]):
    """Schedule repository."""

    model_type = models.Schedule


class SeasonAsyncRepository(SQLAlchemyAsyncSlugRepository[models.Season]):
    """Season repository."""

    model_type = models.Season


class StandingsAsyncRepository(SQLAlchemyAsyncRepository[models.Standings]):
    """Standings repository."""

    model_type = models.Standings


class TeamAsyncRepository(SQLAlchemyAsyncSlugRepository[models.Team]):
    """Team repository."""

    model_type = models.Team


class UserRoleAsyncRepository(SQLAlchemyAsyncRepository[models.UserRole]):
    """UserRole repository."""

    model_type = models.UserRole


class UserAsyncRepository(SQLAlchemyAsyncRepository[models.User]):
    """User repository."""

    model_type = models.User



import os
from functools import lru_cache
from importlib.util import find_spec
from pathlib import Path

from attrs import define, field


@define
class Settings:
    # League Manager app info
    DEFAULT_MODULE_NAME: str = "leaguemanager"
    MODULE_DIR: str = field()

    @MODULE_DIR.default
    def _default_base_dir(self):
        return self._module_to_os_path(self.DEFAULT_MODULE_NAME)

    # User Application directory -
    APP_DIR: Path = field(init=False)

    @APP_DIR.default
    def _default_app_dir(self):
        """Check environment variable for app directory first.

        If not found, return current working directory. This is primarily set if League Manager
        is going to be installed as a dependency.
        """
        try:
            _APP_DIR = os.getenv("APP_DIR")
            if not _APP_DIR:
                return Path.cwd()
            return Path(_APP_DIR)
        except ValueError:
            return Path.cwd()

    # Advanced Alchemy configuration
    ADVANCED_ALCHEMY_CONFIG_DIRNAME: str = field(init=False)

    # ALembic settings
    MIGRATION_PATH: Path = field(init=False)
    MIGRATION_CONFIG: str = field(init=False)
    ALEMBIC_TEMPLATE_PATH: str = field(init=False)

    # SQLite settings
    SQLITE_DATA_DIRECTORY: Path = field(init=False)
    DATE_FORMAT: str = field(default="%Y-%m-%d %H:%M:%S")

    @ADVANCED_ALCHEMY_CONFIG_DIRNAME.default
    def _default_advanced_alchemy_config_dirname(self):
        """Check for environment variable with name of directory containing Advanced Alchemy config object(s).

        If not found, recurse through current directory (ignoring certain directories) to find
        first directory with containing a `.py` file.
        """
        if aa_config_dir := os.getenv("ADVANCED_ALCHEMY_CONFIG_DIRNAME"):
            if not Path(aa_config_dir).is_dir():
                raise ValueError(f"Environment Variable refers to invalid directory: {aa_config_dir}")
            return aa_config_dir
        else:
            _exclude = ["venv", ".venv", "tests", "docs", "migrations"]
            for f in Path.cwd().rglob("**/*.py"):
                match f:
                    case f if any(x in f.parts for x in _exclude):
                        continue
                    case f if f.parent.name == "__pycache__":
                        continue
                    case _:
                        return f.parent.name

    @MIGRATION_PATH.default
    def _default_migration_path(self):
        """Check for environment variable first."""
        if "MIGRATION_PATH" in os.environ:
            migration_path = Path(os.getenv("MIGRATION_PATH"))
            return Path(self.APP_DIR / migration_path, exist_ok=True)
        else:
            return self.APP_DIR / "migrations"

    @MIGRATION_CONFIG.default
    def _default_migration_config(self):
        """If environment variable does not exist, set directory where alembic.ini will be stored."""
        if "MIGRATION_CONFIG" in os.environ:
            return Path(os.getenv("MIGRATION_CONFIG"))
        else:
            return self.APP_DIR

    @ALEMBIC_TEMPLATE_PATH.default
    def _default_alembic_template_path(self):
        """Internally used by League Manager to generate Alembic migration files.

        They are created with the `alembic init` command. However, if the environment variable
        `ALEMBIC_TEMPLATE_PATH` is set, we use that instead for custom Alembic template files."""
        if "ALEMBIC_TEMPLATE_PATH" in os.environ:
            return Path(os.getenv("ALEMBIC_TEMPLATE_PATH"))
        else:
            return Path(self.MODULE_DIR) / "db/alembic_templates"

    @SQLITE_DATA_DIRECTORY.default
    def _default_data_directory(self):
        if "SQLITE_DATA_DIRECTORY" in os.environ:
            return os.getenv("SQLITE_DATA_DIRECTORY")
        return self.APP_DIR / "data_league_db"

    SYNTH_DATA_DIR: Path = field(init=False)

    def __attrs_post_init__(self):
        self.SYNTH_DATA_DIR: Path = Path(self.MODULE_DIR) / "db/synthetic_data"

    def _module_to_os_path(self, module_name: str) -> str:
        """Get the string path of the module."""
        spec = find_spec(module_name)
        if not spec:
            raise ValueError(f"Couldn't find path for {module_name}")
        return str(Path(spec.origin).parent)  # type: ignore[arg-type]


@lru_cache
def get_settings() -> Settings:
    return Settings()



from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from attrs import define, field
from sqlalchemy import select

from leaguemanager.models import League, Season
from leaguemanager.repository import SeasonAsyncRepository, SeasonSyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["SeasonSyncService", "SeasonAsyncService"]


class SeasonSyncService(SQLAlchemySyncRepositoryService):
    """Handles sync database operations for season data."""

    repository_type = SeasonSyncRepository

    def get_leagues_in_season(self, *filters: FilterTypes, season_id: UUID, **kwargs) -> Season:
        """Get the current season."""

        statement = select(League).join_from(Season, League).where(Season.id == season_id)

        return self.list(*filters, statement=statement, **kwargs)

    def get_active_seasons(self, *filters: FilterTypes, **kwargs) -> Season:
        """Get active seasons."""

        return self.list(*filters, **kwargs | {"active": True})


class SeasonAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles async database operations for season data."""

    repository_type = SeasonAsyncRepository

    async def get_leagues_in_season(self, *filters: FilterTypes, season_id: UUID, **kwargs) -> Season:
        """Get the current season."""

        statement = select(League).join_from(Season, League).where(Season.id == season_id)

        return await self.list(*filters, statement=statement, **kwargs)

    async def get_active_seasons(self, *filters: FilterTypes, **kwargs) -> Season:
        """Get active seasons."""

        statement = select(Season).where(Season.active == True)  # noqa

        return await self.list(*filters, statement=statement, **kwargs)



import pytest
from sqlalchemy import delete
from sqlalchemy.ext.asyncio import AsyncSession

from leaguemanager.core.toolbox import clear_table
from leaguemanager.models import League, Season
from leaguemanager.services import LeagueSyncService, SeasonSyncService


def test_create_league(session, league_service) -> None:
    new_league = League(name="Test Create League", category="MEN", division="A", match_day="SUN")
    test_league = league_service.create(new_league)
    assert test_league.name == "Test Create League"
    clear_table(session, League)


def test_create_many_leagues(session, league_service) -> None:
    new_leagues_data = [
        {
            "name": "Test Create League",
            "category": "MEN",
            "division": "A",
            "match_day": "SUN",
        },
        {
            "name": "Test Create League 2",
            "category": "WOMEN",
            "division": "B",
            "match_day": "MON",
        },
    ]

    _ = league_service.create_many(new_leagues_data)

    assert league_service.count() == 2
    clear_table(session, League)


def test_append_league_to_season(session, season_service, league_service) -> None:
    season_data = {
        "name": "SZNZ 2",
        "description": "I Want A Dog",
        "start_date": "2025-01-01",
    }
    league_data = {
        "name": "Songs 2",
        "category": "MEN",
        "division": "A",
        "match_day": "SUN",
    }

    season = season_service.create(data=season_data, auto_commit=True)
    league = league_service.create(data=league_data, auto_commit=True)

    season.leagues.append(league)

    season_service.update(season, auto_commit=True)

    new_season_call = season_service.get_one_or_none(name="SZNZ 2")
    assert new_season_call.leagues[0].name == "Songs 2"

    clear_table(session, League)
    clear_table(session, Season)


def test_delete_league(league_service) -> None:
    new_league = League(name="Test Delete League 2", category="MEN", division="A", match_day="SUN")
    test_league = league_service.create(new_league)
    league_service.delete(test_league.id)
    assert league_service.get_one_or_none(name="Test Delete League 2") is None



import pytest

from leaguemanager.core.toolbox import dynamic_loader


def create_test_files(base_path, files):
    for file_path in files:
        full_path = base_path / file_path
        full_path.parent.mkdir(parents=True, exist_ok=True)
        full_path.touch()


@pytest.fixture
def mock_module_dir(tmp_path_factory):
    base_dir = tmp_path_factory.mktemp("modules_test")
    files = ["module/submodule/__init__.py", "module/submodule/subsubmodule/module1.py", "module/module2.py"]
    create_test_files(base_dir, files)
    return base_dir


# def test_get_modules(mock_module_dir):
#     expected = {"submodule.subsubmodule.module1", "module2"}
#     result = set(get_modules(str(mock_module_dir / "module")))
#     assert result == expected, f"Expected {expected}, got {result}"


@pytest.fixture
def mock_module(mocker):
    """Creates a mock module with a class to be dynamically loaded."""

    class TestClass:
        pass

    # Use mocker to create a mock module
    module = mocker.MagicMock()
    module.TestClass = TestClass
    module.__all__ = ["TestClass"]
    return module


@pytest.fixture
def other_mock_module(mocker):
    """Creates a mock module with a class to be dynamically loaded."""

    class NotTestClass:
        pass

    # Use mocker to create a mock module
    module = mocker.MagicMock()
    module.TestClass = NotTestClass

    return module


def test_dynamic_loader(mock_module, mocker):
    mocker.patch("leaguemanager.core.toolbox.get_modules", return_value=["mock_module", "other_mock_module"])
    mocker.patch.dict("sys.modules", {"mock_module": mock_module, "other_mock_module": mock_module})
    # Mock the compare function to always return True
    compare_func = mocker.MagicMock(return_value=True)

    # Execute dynamic_loader and convert the result to a list to evaluate
    result = list(dynamic_loader("mock_module", compare_func))

    # Assert that our mock class is in the result
    assert mock_module.TestClass in result
    assert mock_module.NotTestClass not in result

    # Additionally, check if compare_func was called with the expected class
    compare_func.assert_called_with(mock_module.TestClass)



from ._async import (
    FixtureAsyncRepository,
    FixtureTeamAsyncRepository,
    LeagueAsyncRepository,
    ManagerAsyncRepository,
    PlayerAsyncRepository,
    RefereeAsyncRepository,
    RoleAsyncRepository,
    ScheduleAsyncRepository,
    SeasonAsyncRepository,
    StandingsAsyncRepository,
    TeamAsyncRepository,
    UserAsyncRepository,
    UserRoleAsyncRepository,
)
from ._sync import (
    FixtureSyncRepository,
    FixtureTeamSyncRepository,
    LeagueSyncRepository,
    ManagerSyncRepository,
    PlayerSyncRepository,
    RefereeSyncRepository,
    RoleSyncRepository,
    ScheduleSyncRepository,
    SeasonSyncRepository,
    StandingsSyncRepository,
    TeamSyncRepository,
    UserRoleSyncRepository,
    UserSyncRepository,
)

__all__ = [
    "FixtureAsyncRepository",
    "FixtureTeamAsyncRepository",
    "LeagueAsyncRepository",
    "ManagerAsyncRepository",
    "PlayerAsyncRepository",
    "RefereeAsyncRepository",
    "RoleAsyncRepository",
    "ScheduleAsyncRepository",
    "SeasonAsyncRepository",
    "StandingsAsyncRepository",
    "TeamAsyncRepository",
    "UserAsyncRepository",
    "UserRoleAsyncRepository",
    "FixtureSyncRepository",
    "FixtureTeamSyncRepository",
    "LeagueSyncRepository",
    "ManagerSyncRepository",
    "PlayerSyncRepository",
    "RefereeSyncRepository",
    "RoleSyncRepository",
    "ScheduleSyncRepository",
    "SeasonSyncRepository",
    "StandingsSyncRepository",
    "TeamSyncRepository",
    "UserRoleSyncRepository",
    "UserSyncRepository",
]



from advanced_alchemy.repository import (
    SQLAlchemySyncRepository,
    SQLAlchemySyncSlugRepository,
)

from leaguemanager import models

__all__ = [
    "FixtureTeamSyncRepository",
    "FixtureSyncRepository",
    "LeagueSyncRepository",
    "ManagerSyncRepository",
    "PlayerSyncRepository",
    "RefereeSyncRepository",
    "RoleSyncRepository",
    "ScheduleSyncRepository",
    "SeasonSyncRepository",
    "TeamSyncRepository",
    "StandingsSyncRepository",
    "UserRoleSyncRepository",
    "UserSyncRepository",
]


class FixtureTeamSyncRepository(SQLAlchemySyncRepository[models.FixtureTeam]):
    """FixtureTeam repository."""

    model_type = models.FixtureTeam


class FixtureSyncRepository(SQLAlchemySyncRepository[models.Fixture]):
    """Fixture repository."""

    model_type = models.Fixture


class LeagueSyncRepository(SQLAlchemySyncSlugRepository[models.League]):
    """League repository."""

    model_type = models.League


class ManagerSyncRepository(SQLAlchemySyncRepository[models.Manager]):
    """Manager repository."""

    model_type = models.Manager


class PlayerSyncRepository(SQLAlchemySyncRepository[models.Player]):
    """Player repository."""

    model_type = models.Player


class RefereeSyncRepository(SQLAlchemySyncRepository[models.Referee]):
    """Referee repository."""

    model_type = models.Referee


class RoleSyncRepository(SQLAlchemySyncSlugRepository[models.Role]):
    """Role repository."""

    model_type = models.Role


class ScheduleSyncRepository(SQLAlchemySyncRepository[models.Schedule]):
    """Schedule repository."""

    model_type = models.Schedule


class SeasonSyncRepository(SQLAlchemySyncSlugRepository[models.Season]):
    """Season repository."""

    model_type = models.Season


class StandingsSyncRepository(SQLAlchemySyncRepository[models.Standings]):
    """Standings repository."""

    model_type = models.Standings


class TeamSyncRepository(SQLAlchemySyncSlugRepository[models.Team]):
    """Team repository."""

    model_type = models.Team


class UserRoleSyncRepository(SQLAlchemySyncRepository[models.UserRole]):
    """UserRole repository."""

    model_type = models.UserRole


class UserSyncRepository(SQLAlchemySyncRepository[models.User]):
    """User repository."""

    model_type = models.User



from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as _UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String, Table
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import UUIDAuditBase, mapper, metadata
from leaguemanager.models.enums import Gender

if TYPE_CHECKING:
    from .team import Team
    from .user import User


@define(slots=False)
class Manager(UUIDAuditBase):
    """A manager of a team."""

    user_id: UUID | None = field(default=None)
    first_name: str | None = field(default=None, validator=validators.max_len(40))
    last_name: str | None = field(default=None, validator=validators.max_len(40))
    verified: bool = field(default=True)
    phone_number: str | None = field(default=None, validator=validators.max_len(20))
    team_email: str | None = field(default=None, validator=validators.max_len(40))
    note: str | None = field(default=None, validator=validators.optional(validators.max_len(255)))


# SQLAlchemy Imperative Mappings

manager = Table(
    "manager",
    metadata,
    Column("id", _UUID, primary_key=True),
    Column("user_id", _UUID, ForeignKey("user.id")),
    Column("first_name", String(40), nullable=False),
    Column("last_name", String(40), nullable=False),
    Column("verified", Boolean),
    Column("phone_number", String(20)),
    Column("team_email", String(40)),
    Column("note", String(255)),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)

# ORM Relationships

mapper.map_imperatively(
    Manager,
    manager,
    properties={
        "user": relationship("User", back_populates="manager"),
        "teams": relationship("Team", back_populates="manager"),
    },
)



from datetime import UTC, datetime, timedelta
from typing import Any
from uuid import UUID

from advanced_alchemy.filters import CollectionFilter, OnBeforeAfter
from sqlalchemy import select

# from league_manager.domain.scheduling import ScheduleGenerator
from leaguemanager.models import Fixture, FixtureTeam, Team
from leaguemanager.repository import (
    FixtureAsyncRepository,
    FixtureSyncRepository,
    FixtureTeamAsyncRepository,
    FixtureTeamSyncRepository,
)
from leaguemanager.services._typing import ModelT
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = [
    "FixtureSyncService",
    "FixtureTeamSyncService",
    "FixtureAsyncService",
    "FixtureTeamAsyncService",
]


class FixtureSyncService(SQLAlchemySyncRepositoryService):
    """Handles sync database operations for fixtures."""

    repository_type = FixtureSyncRepository

    def all_fixtures_in_schedule(self, schedule_id: UUID) -> list[Fixture]:
        statement = select(self.model_type).where(self.model_type.schedule_id == schedule_id)
        return self.list(statement=statement)

    def past_fixtures(self, schedule_id: UUID) -> list[Fixture]:
        filter = CollectionFilter(field_name="status", values=["P", "F", "A"])
        statement = select(self.model_type).where((self.model_type.schedule_id == schedule_id))
        return self.list(filter, statement=statement)

    def remaining_fixtures(self, schedule_id: UUID) -> list[Fixture]:
        filter = CollectionFilter(field_name="status", values=["U", "D"])
        statement = select(self.model_type).where((self.model_type.schedule_id == schedule_id))
        return self.list(filter, statement=statement)

    def next_scheduled_fixtures(
        self, schedule_id: UUID, *, days: int = 6, filter: OnBeforeAfter | None = None
    ) -> list[Fixture]:
        if filter is None:
            filter = OnBeforeAfter(
                field_name="date",
                on_or_after=datetime.now(UTC),
                on_or_before=datetime.now(UTC) + timedelta(days=days),
            )
        statement = select(self.model_type).where((self.model_type.schedule_id == schedule_id))
        return self.list(filter, statement=statement)


class FixtureTeamSyncService(SQLAlchemySyncRepositoryService):
    """Handles sync database operations for fixture_team association data."""

    repository_type = FixtureTeamSyncRepository


class FixtureAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles async database operations for fixtures."""

    repository_type = FixtureAsyncRepository


class FixtureTeamAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles async database operations for fixture_team association data."""

    repository_type = FixtureTeamAsyncRepository



from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import Standings, Team
from leaguemanager.repository import StandingsAsyncRepository, StandingsSyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["StandingsSyncService", "StandingsAsyncService"]


class StandingsSyncService(SQLAlchemySyncRepositoryService):
    """Handles sync database operations for standings."""

    repository_type = StandingsSyncRepository


class StandingsAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles async database operations for standings."""

    repository_type = StandingsAsyncRepository



from __future__ import annotations

from datetime import UTC, datetime
from typing import TYPE_CHECKING, Any
from uuid import UUID, uuid4

from attrs import define, field
from sqlalchemy import Column, ForeignKey, MetaData, Table
from sqlalchemy.orm import Mapper, registry
from sqlalchemy.types import UUID as _UUID

mapper = registry()
metadata = MetaData()


@define(slots=False)
class UUIDBase:
    id: UUID = field(factory=uuid4)

    if TYPE_CHECKING:
        __name__: str
        __mapper__: Mapper[Any]

    def to_dict(self, exclude: set[str] | None = None) -> dict[str, Any]:
        """Convert model to dictionary.

        Returns:
            Dict[str, Any]: A dict representation of the model
        """
        exclude = {"sa_orm_sentinel", "_sentinel"}.union(self._sa_instance_state.unloaded).union(exclude or [])  # type: ignore[attr-defined]
        return {
            field: getattr(self, field)
            for field in self.__mapper__.columns.keys()  # noqa: SIM118
            if field not in exclude
        }


@define(slots=False)
class UUIDAuditBase(UUIDBase):
    created_at: datetime = field(factory=lambda: datetime.now(UTC))
    updated_at: datetime = field(factory=lambda: datetime.now(UTC))



from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as _UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String, Table, UniqueConstraint
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.core import get_settings
from leaguemanager.core.toolbox import str_to_iso
from leaguemanager.models.base import UUIDAuditBase, mapper, metadata

if TYPE_CHECKING:
    from .fixture import Fixture
    from .league import League
    from .season import Season

settings = get_settings()


@define(slots=False)
class Schedule(UUIDAuditBase):
    """Defines the schedule for a league and rules for fixtures."""

    league_id: UUID | None = field(default=None)
    season_id: UUID | None = field(default=None)
    name: str | None = field(default=None, validator=validators.optional(validators.max_len(80)))
    total_games: int = field(default=10)
    game_length: int = field(default=90)
    half_time: int = field(default=15)
    time_between_games: int = field(default=15)
    concurrent_games: int = field(default=2)
    start_date: datetime | None = field(default=None, converter=lambda x: str_to_iso(x, settings.DATE_FORMAT))
    end_date: datetime | None = field(default=None)
    active: bool = field(default=True)

    # season_name: AssociationProxy[str] = association_proxy("season", "name")


# SQLAlchemy Imperative Mappings

schedule = Table(
    "schedule",
    metadata,
    Column("id", _UUID, primary_key=True),
    Column("league_id", _UUID, ForeignKey("league.id"), nullable=True),
    Column("season_id", _UUID, ForeignKey("season.id"), nullable=True),
    UniqueConstraint("id", "league_id", "season_id", name="uxschedule"),
    Column("name", String(80), nullable=True),
    Column("total_games", Integer, nullable=False),
    Column("game_length", Integer, nullable=False),
    Column("half_time", Integer, nullable=False),
    Column("time_between_games", Integer, nullable=False),
    Column("concurrent_games", Integer, nullable=False),
    Column("start_date", DateTime(20), nullable=True),
    Column("end_date", DateTime(20), nullable=True),
    Column("active", Boolean, nullable=False),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)

# ORM Relationships

mapper.map_imperatively(
    Schedule,
    schedule,
    properties={
        "league": relationship("League", back_populates="schedule", innerjoin=True, uselist=False, lazy="joined"),
        "season": relationship("Season", back_populates="schedules"),
        "fixtures": relationship("Fixture", back_populates="schedule", uselist=True),
        "team_standings": relationship("Standings", back_populates="schedule"),
    },
)



from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import Player, User
from leaguemanager.repository import PlayerAsyncRepository, PlayerSyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["PlayerSyncService", "PlayerAsyncService"]


class PlayerSyncService(SQLAlchemySyncRepositoryService):
    """Handles sync database operations for players."""

    repository_type = PlayerSyncRepository


class PlayerAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles async database operations for players."""

    repository_type = PlayerAsyncRepository



from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import Schedule, Season, Team
from leaguemanager.repository import ScheduleAsyncRepository, ScheduleSyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["ScheduleSyncService", "ScheduleAsyncService"]


class ScheduleSyncService(SQLAlchemySyncRepositoryService):
    """Handles sync database operations for scheduling."""

    repository_type = ScheduleSyncRepository


class ScheduleAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles async database operations for scheduling."""

    repository_type = ScheduleAsyncRepository



from __future__ import annotations

from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as _UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import UUIDAuditBase, mapper, metadata
from leaguemanager.models.enums import Field, FixtureStatus

from ._association import fixture_referee

if TYPE_CHECKING:
    from .fixture_team import FixtureTeam
    from .league import League
    from .player import Player
    from .referee import Referee
    from .schedule import Schedule
    from .season import Season
    from .team import Team


@define(slots=False)
class Fixture(UUIDAuditBase):
    """Represents a specific match in a schedule."""

    season_id: UUID | None = field(default=None)
    league_id: UUID | None = field(default=None)
    schedule_id: UUID | None = field(default=None)

    home_team_id: UUID | None = field(default=None)
    away_team_id: UUID | None = field(default=None)

    bye_week: bool = field(default=False)

    home_goals: int | None = field(default=None)
    away_goals: int | None = field(default=None)

    matchday: int | None = field(default=None)
    date: str | None = field(default=None)
    status: FixtureStatus = field(
        default=FixtureStatus.U.name, validator=validators.in_([v.name for v in FixtureStatus])
    )
    field: Field = field(default=Field.A.name, validator=validators.in_([v.name for v in Field]))


# SQLAlchemy Imperative Mappings

fixture = Table(
    "fixture",
    metadata,
    Column("id", _UUID, primary_key=True),
    Column("season_id", _UUID, ForeignKey("season.id")),
    Column("league_id", _UUID, ForeignKey("league.id")),
    Column("schedule_id", _UUID, ForeignKey("schedule.id", ondelete="cascade")),
    Column("home_team_id", _UUID, ForeignKey("team.id", ondelete="set null")),
    Column("away_team_id", _UUID, ForeignKey("team.id", ondelete="set null")),
    Column("bye_week", Boolean),
    Column("home_goals", Integer),
    Column("away_goals", Integer),
    Column("matchday", Integer),
    Column("date", DateTime),
    Column("field", String(2)),
    Column("status", String(10)),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)


# ORM Relationships

mapper.map_imperatively(
    Fixture,
    fixture,
    properties={
        "schedule": relationship("Schedule", back_populates="fixtures"),
        "referees": relationship("Referee", secondary=fixture_referee, back_populates="fixtures"),
        "teams": relationship("FixtureTeam", back_populates="fixture", cascade="all, delete"),
    },
)



from __future__ import annotations

from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as _UUID
from sqlalchemy import Boolean, Column, ForeignKey, Integer, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import UUIDBase, mapper, metadata
from leaguemanager.models.enums import FixtureResult

if TYPE_CHECKING:
    from .fixture import Fixture
    from .team import Team


@define(slots=False)
class FixtureTeam(UUIDBase):
    """Results for a fixture for specific team (association table)."""

    team_id: UUID | None = field(default=None)
    fixture_id: UUID | None = field(default=None)

    is_played: bool = field(default=False)
    fixture_result: FixtureResult | None = field(
        default=None, validator=validators.optional(validators.in_([v.name for v in FixtureResult]))
    )
    points: int | None = field(default=0)
    goals_for: int | None = field(default=0)
    goals_against: int | None = field(default=0)


# SQLAlchemy Imperative Mappings

fixture_team = Table(
    "fixture_team",
    metadata,
    Column("id", _UUID, primary_key=True),
    Column("team_id", _UUID, ForeignKey("team.id")),
    Column("fixture_id", _UUID, ForeignKey("fixture.id")),
    Column("is_played", Boolean),
    Column("fixture_result", String(10)),
    Column("points", Integer),
    Column("goals_for", Integer),
    Column("goals_against", Integer),
)

# ORM Relationships

mapper.map_imperatively(
    FixtureTeam,
    fixture_team,
    properties={
        "team": relationship("Team", back_populates="fixtures"),
        "fixture": relationship("Fixture", back_populates="teams"),
    },
)



from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as _UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String, Table
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import UUIDAuditBase, mapper, metadata
from leaguemanager.models.enums import Gender

from ._association import player_team

if TYPE_CHECKING:
    from ..account.user import User
    from .team import Team


@define(slots=False)
class Player(UUIDAuditBase):
    """A player in a team."""

    first_name: str | None = field(default=None, validator=validators.max_len(40))
    last_name: str | None = field(default=None, validator=validators.max_len(40))
    age_verified: bool = field(default=False)
    age: int | None = field(default=None)
    gender: str | None = field(default=None, validator=validators.optional(validators.in_([v.name for v in Gender])))
    note: str | None = field(default=None, validator=validators.optional(validators.max_len(255)))


# SQLAlchemy Imperative Mappings

player = Table(
    "player",
    metadata,
    Column("id", _UUID, primary_key=True),
    Column("first_name", String(40), nullable=False),
    Column("last_name", String(40), nullable=False),
    Column("age", Integer),
    Column("age_verified", Boolean),
    Column(
        "gender",
        String(2),
    ),
    Column("note", String(255)),
    Column("user_profile_id", _UUID, ForeignKey("user.id")),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)

# ORM Relationships

mapper.map_imperatively(
    Player,
    player,
    properties={
        "user": relationship("User", back_populates="player"),
        "teams": relationship("Team", secondary=player_team, back_populates="players"),
    },
)



from sqlalchemy import delete

from leaguemanager.core.toolbox import clear_table
from leaguemanager.models import League, Season
from leaguemanager.services import LeagueSyncService, SeasonSyncService


def test_create_many_season(session, season_service) -> Season:
    # clear_table(session, Season)
    season_data = [
        {"name": "Test Create Season", "description": "Test Description", "start_date": "2022-01-01"},
        {"name": "Test Create Season 2", "description": "Test Description 2", "start_date": "2022-01-02"},
        {"name": "Test Create Season 3", "description": "Test Description 3", "start_date": "2022-01-03"},
    ]

    seasons = season_service.create_many(season_data)

    assert seasons[0].name == "Test Create Season"
    assert seasons[1].name == "Test Create Season 2"
    assert seasons[2].name == "Test Create Season 3"

    clear_table(session, Season)


def test_attach_season_to_league(session, season_service, league_service) -> None:
    season_data = {
        "name": "SZNZ",
        "description": "I Want A Dog",
        "start_date": "2025-01-01",
    }
    league_data = {
        "name": "Songs",
        "category": "MEN",
        "division": "A",
        "match_day": "SUN",
    }

    season = season_service.create(data=season_data, auto_commit=True)
    league = league_service.create(data=league_data, auto_commit=True)

    league = league_service.get_one_or_none(name="Songs")

    league.season = season

    league_service.update(league, auto_commit=True)

    new_league_call = league_service.get_one_or_none(name="Songs")
    assert new_league_call.season.name == "SZNZ"
    assert season.leagues[0].name == "Songs"

    clear_table(session, League)
    clear_table(session, Season)


def test_select_only_active_seasons(session, season_service) -> None:
    active_season = Season(
        name="Active Season-test",
        description="I Want A Dog",
        start_date="2025-01-01",
        active=True,
    )

    inactive_season = Season(
        name="Inactive Season-test",
        description="I Want A Dog",
        start_date="2025-01-01",
        active=False,
    )

    season_service.create_many([active_season, inactive_season])

    seasons = season_service.get_active_seasons()

    assert len(seasons) == 1
    assert seasons[0].name == "Active Season-test"

    clear_table(session, Season)



"""Metadata for CLI tool."""

__app_name__ = "mgr"


from leaguemanager.__metadata__ import __version__
from leaguemanager.domain import Scheduler, StandingsTable, TabulateFixture

# from .domain.scheduling import _generator

__all__ = ["__app_name__", "__version__", "Scheduler", "StandingsTable", "TabulateFixture"]



import inspect
from datetime import datetime
from importlib import import_module
from inspect import isclass, signature
from pathlib import Path
from typing import Any, Iterable, Iterator

from advanced_alchemy.config import SQLAlchemyAsyncConfig, SQLAlchemySyncConfig
from advanced_alchemy.repository import SQLAlchemyAsyncRepository, SQLAlchemySyncRepository
from advanced_alchemy.service import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService
from sqlalchemy import delete
from sqlalchemy.orm import Session

from leaguemanager.core import get_settings
from leaguemanager.models.base import UUIDBase

settings = get_settings()


def get_modules(module: str) -> Iterator[str]:
    """Returns all .py modules in given file_dir.

    Args:
        module (str): Name of directory to begin recursive search

    Yields:
        Iterator[str]: The generator contains paths to modules
        with dot notation starting at project root.
        (For example: "app.models.user")

    References:
        [Bob Waycott](
            https://bobwaycott.com/blog/how-i-use-flask/organizing-flask-models-with-automatic-discovery/
            )
    """
    app_root = Path(settings.MODULE_DIR)
    if app_root.name == module:
        file_dir = app_root
    else:
        file_dir = app_root / module
    idx_app_root = len(app_root.parts) - 1  # index of app root
    modules = [f for f in list(file_dir.rglob("*.py")) if not f.stem == "__init__"]
    for filepath in modules:
        yield (".".join(filepath.parts[idx_app_root:])[0:-3])


def dynamic_loader(module: str, compare: bool) -> Iterable:
    """Iterates over all .py files in `module` directory, finding all classes that
    match `compare` function.

    Other classes/objects in the module directory will be ignored. Note that only objects
    contained within the `__all__` attribute will be loaded.

    Returns unique items found.

    Args:
        module (str): Directory name to search recursively
        compare (function): Boolean comparison of all py files in
        `module` directory

    Returns:
        list: All modules that match the `compare` function.

    References:
        [Bob Waycott](
            https://bobwaycott.com/blog/how-i-use-flask/organizing-flask-models-with-automatic-discovery/
            )
    """
    items = []
    for mod in get_modules(module):
        try:
            _module = import_module(mod)
        except AttributeError:
            # Ignore modules that create circular imports
            continue
        if hasattr(_module, "__all__"):
            objs = [getattr(_module, obj) for obj in _module.__all__]
            items += [o for o in objs if compare(o) and o not in items]

    return items


def get_signature(func):
    """Returns signature of given function"""
    if hasattr(func, "__wrapped__"):
        return signature(func.__wrapped__)
    return signature(func)


def is_sync_repo(item: Any):
    """Checks if item is SQLAlchemySyncRepository class or subclass"""
    return isclass(item) and issubclass(item, SQLAlchemySyncRepository)


def is_async_repo(item: Any):
    """Checks if item is SQLAlchemyAsyncRepository class or subclass"""
    return isclass(item) and issubclass(item, SQLAlchemyAsyncRepository)


def is_sync_service(item: Any):
    """Checks if item is SQLAlchemySyncRepositoryService class or subclass."""
    return isclass(item) and issubclass(item, SQLAlchemySyncRepositoryService)


def is_async_service(item: Any):
    """Checks if item is SQLAlchemyAsyncRepositoryService class or subclass."""
    return isclass(item) and issubclass(item, SQLAlchemyAsyncRepositoryService)


def is_sync_config(item: Any):
    """Checks if item is a SQLAlchemySyncConfig.

    Was unable to use `issubclass` here. May need to look into it more at some point.
    """
    return item.__class__ == SQLAlchemySyncConfig


def is_async_config(item: Any):
    """Checks if item is a SQLAlchemyAsyncConfig.

    Was unable to use `issubclass` here. May need to look into it more at some point.
    """
    return item.__class__ == SQLAlchemyAsyncConfig


def get_services(is_async: bool = False) -> list[SQLAlchemySyncRepositoryService | SQLAlchemyAsyncRepositoryService]:
    """Returns all SQLAlchemySyncRepositoryService or SQLAlchemyAsyncRepositoryService classes dynamically.

    Args:
        is_async (bool, optional): If True, returns all SQLAlchemyAsyncRepositoryService classes.
        Defaults to False.
    """
    if is_async:
        return dynamic_loader("services", is_async_service)
    return dynamic_loader("services", is_sync_service)


def get_repositories(is_async: bool = False) -> list[SQLAlchemyAsyncRepository | SQLAlchemySyncRepository]:
    """Returns all SQLAlchemySyncRepository or SQLAlchemyAsyncRepository classes dynamically.

    Args:
        is_async (bool, optional): If True, returns all SQLAlchemyAsyncRepository classes.
        Defaults to False.

    Returns:
        list: List of SQLAlchemySyncRepository or SQLAlchemyAsyncRepository classes
    """
    if is_async:
        return dynamic_loader("repository", is_async_repo)
    return dynamic_loader("repository", is_sync_repo)


def get_advanced_alchemy_config(module_name: str = "app", is_async: bool = False):
    """Returns Advanced Alchemy config classes, including SQLAlchemySyncConfig and SQLAlchemyAsyncConfig.

    This function will find any `SQLAlchemySyncConfig` or `SQLAlchemyAsyncConfig` classes (or subclasses) contained
    within the `module_name` directory. It will search recursively and return all classes found. However, the
    module must be included in the `__all__` list of wherever it is imported. For example:

    Example:
        >>> from advanced_alchemy.config import SQLAlchemySyncConfig
        >>>
        >>> __all__ = ["sync_config"]
        >>>
        >>> sync_config = SQLAlchemySyncConfig(...)

    If the `module_name` is `db_config`, then it will search the `db_config` directory and all subdirectories for
    a matching `SQLAlchemySyncConfig` or `SQLAlchemyAsyncConfig` class.

    If no `module_name` is provided, it will default to "app", and search there recursively.

    If none are found, it will return an empty list.

    Args:
        module_name (str, optional): The name of the module to search within. Defaults to "app".
        is_async (bool, optional): If True, returns all SQLAlchemyAsyncConfig classes. Defaults to False.

    Returns:
        list: A list of SQLAlchemySyncConfig or SQLAlchemyAsyncConfig classes.
    """
    if is_async:
        return dynamic_loader(module_name, is_async_config)
    return dynamic_loader(module_name, is_sync_config)


def clear_table(session: Session, model: UUIDBase) -> None:
    """Clears table of given model."""
    session.execute(delete(model))


def str_to_iso(date_string: str, format: str):
    """Converts string to datetime object."""
    return datetime.strptime(date_string, format)



from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String, Table, UniqueConstraint
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import UUIDAuditBase, mapper, metadata
from leaguemanager.models.enums import Category, Division, MatchDay

if TYPE_CHECKING:
    from .schedule import Schedule
    from .season import Season
    from .team import Team


@define(slots=False)
class League(UUIDAuditBase):
    """Defines a league with certain rules, as well as divisions and categories.

    TODO: Add a way to create custom ruleset that applies to league."""

    season_id: UUID | None = field(default=None)

    active: bool = field(default=True)
    category: Category | None = field(
        default=None, validator=validators.optional(validators.in_([v.name for v in Category]))
    )
    division: Division | None = field(
        default=None, validator=validators.optional(validators.in_([v.name for v in Division]))
    )
    match_day: MatchDay | None = field(
        default=None, validator=validators.optional(validators.in_([v.name for v in MatchDay]))
    )

    name: str | None = field(default=None, validator=validators.max_len(80))
    team_size: int | None = field(default=11)

    age_rule: bool = field(default=False)
    age_rule_min: int | None = field(default=None)
    age_rule_max: int | None = field(default=None)

    m_age_rule_min: int | None = field(default=None)
    m_age_rule_max: int | None = field(default=None)
    f_age_rule_min: int | None = field(default=None)
    f_age_rule_max: int | None = field(default=None)

    # season: "Season" | None = field(default=None)


# SQLAlchemy Imperative Mapping

league = Table(
    "league",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("season_id", SA_UUID, ForeignKey("season.id")),
    Column("name", String(80), nullable=False, unique=True),
    Column("category", String(20), nullable=True),
    Column("division", String(20), nullable=True),
    Column("match_day", String(20), nullable=True),
    Column("team_size", Integer, nullable=True),
    Column("active", Boolean, nullable=False),
    Column("age_rule", Boolean, nullable=False),
    Column("age_rule_min", Integer, nullable=True),
    Column("age_rule_max", Integer, nullable=True),
    Column("m_age_rule_min", Integer, nullable=True),
    Column("m_age_rule_max", Integer, nullable=True),
    Column("f_age_rule_min", Integer, nullable=True),
    Column("f_age_rule_max", Integer, nullable=True),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)

# ORM Relationships

mapper.map_imperatively(
    League,
    league,
    properties={
        "season": relationship("Season", back_populates="leagues", lazy="selectin"),
        "schedule": relationship("Schedule", back_populates="league", uselist=False, lazy="selectin"),
        "teams": relationship("Team", back_populates="league", lazy="selectin"),
    },
)



from sqlalchemy import Column, ForeignKey, MetaData, Table
from sqlalchemy.types import UUID as _UUID

from ..base import metadata

# Association tables

player_team = Table(
    "player_team",
    metadata,
    Column("player_id", _UUID, ForeignKey("player.id"), primary_key=True),
    Column("team_id", _UUID, ForeignKey("team.id"), primary_key=True),
)

fixture_referee = Table(
    "fixture_referees",
    metadata,
    Column("fixture_id", _UUID, ForeignKey("fixture.id"), primary_key=True),
    Column("referee_id", _UUID, ForeignKey("referee.id"), primary_key=True),
)
